;**********************************************************************
;
; Tutorial 12.2R - Manchester receive routine, four byte packets.
; display via RS232 at 9600 baud
;
;**********************************************************************

	list      p=16f628

	__CONFIG 0x3D18

	#include <p16f628.inc>


;***** CONSTANT VALUES
EXPIRE_TIMER	EQU 		0x12
LATCH_MASK	EQU 		0xff
VALID_BIT	EQU 		4
hdrcntmin	EQU		0x0c ; minimum number of header bits to receive
hdrcntmax	EQU		0x10 ; maximum number of header bits to receive

; decoder time tolerances are set at : 0.5T, 1.5T, 2.5T
; measured in 9xinstr.time (9usec) counts

T		EQU		.39 ; half frame 350 usec (= T * 9 usec)

min_t		EQU		T/2  ; half frame (T) minimum time
min_2t		EQU		3*T/2 ; half frame (T) maximum time and full frame (2T) minimum time
max_2t		EQU		5*T/2 ; full frame (2T) maximum time

packet_len	EQU		5 ; packet length, check var. alloc!
mrx_packet_len	EQU		5 ; packet length, check var. alloc!


; input port bit
#define 	RXBIT 		PORTA, 7

; normal decoder logic input
#define 	SKL 		btfsc
#define 	SKH 		btfss

; inverse decoder logic input
;#define 	SKL 		btfss
;#define 	SKH 		btfsc

;***** FLAGS
#define 	IF_SHORT 	flags, 0
#define 	FIRST_HALF 	flags, 1
#define 	HEADER 		flags, 2
#define 	VALID 		flags, 7

if_short_val	EQU		1 ; bit value of IF_SHORT flag
first_half_val	EQU		2 ; bit value of FIRST_HALF flag


SER_PORT	Equ	PORTB
SER_TRIS	Equ	TRISB
SER_IN		Equ	0x07
SER_OUT		Equ	0x06

;***** VARIABLE DEFINITIONS

	cblock 	0x20 			;start of general purpose registers

		bitcnt		
		tmrval		 	; timer value
		bt		 	; receive byte buffer
		flags		 	; decode logic status
		btcnt		 	; byte counter

		mrx_buffer 		; receive address buffer
		mrx_buffer1		; receive packet buffer
		mtx_buffer2
		mtx_buffer3
		mtx_buffer4
		mrx_bsum		; receive buffer, checksum

		count			;used in looping routines
		count1			;used in delay routine
		counta			;used in delay routine
		countb			;used in delay routine
		tmp1			;temporary storage
		tmp2
		Xmit_Byte        	;holds byte to xmit
            	Rcv_Byte        	;holds received byte 
            	Bit_Cntr        	;bit counter for RS232
            	Delay_Count        	;delay loop counter

	endc

		org	0x0000		; org sets the origin, 0x0000 for the 16F628,
  		goto    main		; go to beginning of program
		nop
		nop
		nop
					; interrupt vector location
		retfie

main
		;HARDWARE INIT
		movlw 	7
		movwf 	CMCON

SetPorts	bsf 	STATUS,		RP0	; select bank 1
		movlw 	0x80 			; A7=input
		movwf 	TRISA
		bcf 	STATUS,		RP0	; select bank 0
		call	SER_INIT		; initialise serial port

loop		

		call 	mrx_receive		; get packet
		andlw 	0xff
		bnz 	loop			; reject if error code
		call 	mrx_chk_buf		; test the checksum value
		andlw 	0xff
		bnz 	loop			; reject packet if checksum fails

		movlw	0xAA			; check that packet address is correct (0xAA)
		subwf 	(mrx_buffer), W
		bnz 	loop			; else reject packet

		movf	mrx_buffer+1, W		; get data byte byte
		call	XMIT_RS232		; and display it
		movf	mrx_buffer+2, W		; get data byte byte
		call	XMIT_RS232		; and display it
		movf	mrx_buffer+3, W		; get data byte byte
		call	XMIT_RS232		; and display it
		movf	mrx_buffer+4, W		; get data byte byte
		call	XMIT_RS232		; and display it

		goto 	loop			; just display the bytes

;*************** receive routines *************************

;    Author:        el@jap.hu
;                   http://jap.hu/electronic/


mrx_receive	; receive a full manchester-encoded packet
s3		; set flags: first_half=1, if_short=0
		bsf 	FIRST_HALF
s4		bcf 	IF_SHORT

s5		; init before the received packet

		; set FSR to buffer start
		movlw 	mrx_buffer
		movwf 	FSR
		; set byte counter
		movlw 	(packet_len+1) ; bytes / packet
		movwf 	btcnt
		; set header receive mode
		bsf 	HEADER
		clrf 	bitcnt ; counting bit1-s in this mode

s2		; wait for a pulse
		SKH 	RXBIT
		goto 	s2

s6		; wait for end of (short) pulse up to min_2t
		clrf 	tmrval
s6_w		SKH 	RXBIT
		goto 	s7 		; goto s7 at end of pulse

		incf 	tmrval, F
		nop
		movlw 	min_2t
		subwf 	tmrval, W
		btfss 	STATUS, C
		goto 	s6_w

		; timeout, exit
		retlw 	1 		; illegal startbit

s7		; start timer
		clrf 	tmrval

s8		; if (if_short & rxbit) goto s9
		; if (!if_short & !rxbit) goto s9
		; goto s10

		btfsc 	IF_SHORT
		; if_short = 1
		goto 	s8_ss1

s8_ss0		; if_short = 0
		SKL 	RXBIT
		goto 	s10 		; rxbit = 1, goto s10

s9_ss0		; if (timer > max_2t) exit - else goto s8
		movlw 	max_2t
		subwf 	tmrval, W
		btfsc 	STATUS, C
		retlw 	2 		; signal too long

		incf 	tmrval, F
		goto 	s8_ss0

s8_ss1		; if_short = 1
		SKH 	RXBIT
		goto 	s10 ; rxbit = 0, goto s10

s9_ss1		; if (timer > max_2t) exit - else goto s8
		movlw 	max_2t
		subwf 	tmrval, W
		btfsc 	STATUS, C
		retlw 	2 		; signal too long

		incf 	tmrval, F
		goto 	s8_ss1

s10		; invert if_short
		movlw 	if_short_val
		xorwf 	flags, F

s11		; if (timer < min_t) exit
		movlw 	min_t
		subwf 	tmrval, W
		btfss 	STATUS, C
		retlw 3 		; signal too short

s12		; if (timer < min_2t) goto s14
		movlw 	min_2t
		subwf 	tmrval, W
		btfss 	STATUS, C
		goto 	s14

s13		; if (first_half = 0) goto s16 - else exit
		btfss 	FIRST_HALF
		goto 	s16
		retlw 	4 		; no mid-frame transition/out of sync

s14		; invert first_half
		movlw 	first_half_val
		xorwf 	flags, F

s15		; if (first_half = 1) goto 7
		btfsc 	FIRST_HALF
		goto 	s7

s16		; if_short is a decoded bit. Handle here
		btfss 	HEADER
		goto 	s16_not_header

		; header receiving mode
		btfss 	IF_SHORT
		goto 	s16_header_end

		; header bit is 1
		btfss 	bitcnt, 4 ; inc up to 16
		incf 	bitcnt, F  ; 16 is enough...

#ifdef NOMAXHDR
		; test for max header length
		movlw 	hdrcntmax
		subwf 	bitcnt, W
		btfss 	STATUS, C
#endif
		goto 	s7 ; loop back
		retlw 	9 		; header too long

s16_header_end	; header ends indicated by a 0
		bcf 	HEADER

		; test for min header length
		movlw 	hdrcntmin
		subwf 	bitcnt, W
		btfss 	STATUS, C
		retlw 	0x0a 		; header too short

next_byte	movlw 	0x0a
		movwf 	bitcnt
		goto 	s7 ; loop back

s16_not_header	; receiving bytes
		decf 	bitcnt, F
		bz 	s16_s4 ; if (bitcnt = 0) check for a byte-sep 1

		; if (bitcnt = 1) check for a byte-separator 0
		movlw 	1
		xorwf 	bitcnt, W
		bnz 	s16_s2

		; test for a byte separator 1
		btfsc 	IF_SHORT
		goto 	s7
		retlw 	7 		; byte-ending 1 not present

s16_s2		; bit is data

		rrf 	flags, W
		rlf 	bt, F
		goto 	s7

s16_s4		; check for a byte-separator 0
		btfsc 	IF_SHORT
		retlw 	8 		; byte-ending 0 not present

		; OK, received byte is sane, store in buffer
		movf 	bt, W
		movwf 	INDF
		incf 	FSR, F

		decfsz 	btcnt, F
		goto 	next_byte

		retlw 	0 		; OK, buffer received

		; buffer checking is not done automatically!
		; if returned value is 0, call mrx_chk_buf to check
		;

mrx_chk_buf	; check buffer sanity by chksum
		movlw 	mrx_buffer
		movwf 	FSR
		movlw 	(packet_len+1) ; number of bytes with the chksum byte
		movwf 	btcnt
		movlw 	0xff
		movwf 	bt ; used as sum register

chk0		movf 	INDF, W
		; fast CRC-8 algorithm with poly x^8+x^5+x^4+1
		; executes in 23 cycles per update
		xorwf	bt,f
		clrw
		btfsc	bt,7
		xorlw	0x7a
		btfsc	bt,6
		xorlw	0x3d
		btfsc	bt,5
		xorlw	0x86
		btfsc	bt,4
		xorlw	0x43
		btfsc	bt,3
		xorlw	0xb9
		btfsc	bt,2
		xorlw	0xc4
		btfsc	bt,1
		xorlw	0x62
		btfsc	bt,0
		xorlw	0x31
		movwf	bt

		incf 	FSR, F
		decfsz 	btcnt, F
		goto 	chk0
		; correct checksum must be zero
		movf 	bt, W
		bnz 	chk_err
		retlw 	0 		; result is in Z
chk_err		retlw 	0x0c 		; checksum error

mrx_init	return

;delay routines

Delay255	movlw	0xff			;delay 255 mS
		goto	d0
Delay100	movlw	d'100'			;delay 100mS
		goto	d0
Delay50		movlw	d'50'			;delay 50mS
		goto	d0
Delay20		movlw	d'20'			;delay 20mS
		goto	d0
Delay5		movlw	0x05			;delay 5.000 ms (4 MHz clock)
d0		movwf	count1
d1		movlw	0xC7			;delay 1mS
		movwf	counta
		movlw	0x01
		movwf	countb
Delay_0
		decfsz	counta, f
		goto	$+2
		decfsz	countb, f
		goto	Delay_0

		decfsz	count1	,f
		goto	d1
		retlw	0x00

;Serial routines

SER_INIT
            BSF     STATUS, RP0           ;select bank 1
            BCF     SER_TRIS, SER_OUT     ;set B6 as an output
            BSF     SER_TRIS, SER_IN      ;set B7 as an input
            BCF     STATUS, RP0           ;select bank 0
            BSF     SER_PORT, SER_OUT     ;set SER_OUT high
            RETURN

XMIT_RS232  MOVWF   Xmit_Byte             ;move W to Xmit_Byte
            MOVLW   0x08                  ;set 8 bits out
            MOVWF   Bit_Cntr
            BCF     SER_PORT, SER_OUT
            CALL    Bit_Delay
Ser_Loop    RRF     Xmit_Byte , f         ;send one bit
            BTFSS   STATUS    , C
            BCF     SER_PORT, SER_OUT
            BTFSC   STATUS    , C
            BSF     SER_PORT, SER_OUT
            CALL    Bit_Delay
            DECFSZ  Bit_Cntr  , f         ;test if all done
            GOTO    Ser_Loop
            BSF     SER_PORT, SER_OUT
            CALL    Bit_Delay
            RETURN

Rcv_RS232   BTFSC   SER_PORT, SER_IN      ;wait for start bit
            GOTO    Rcv_RS232
            CALL    Start_Delay	          ;do half bit time delay
            BTFSC   SER_PORT, SER_IN      ;check still in start bit
            GOTO    Rcv_RS232
            MOVLW   0x08                  ;set up to read 8 bits
            MOVWF   Bit_Cntr
            CLRF    Rcv_Byte
Next_RcvBit CALL    Bit_Delay
            BTFSS   SER_PORT, SER_IN
            BCF     STATUS    , C
            BTFSC   SER_PORT, SER_IN
            BSF     STATUS    , C
            RRF     Rcv_Byte  , f
            DECFSZ  Bit_Cntr  , f         ;test if all done
            GOTO    Next_RcvBit
            CALL    Bit_Delay
            MOVF    Rcv_Byte, W
            RETURN

Start_Delay MOVLW   0x0C
            MOVWF   Delay_Count
Start_Wait  NOP
            DECFSZ  Delay_Count , f
            GOTO    Start_Wait
            RETURN

Bit_Delay   MOVLW   0x18
            MOVWF   Delay_Count
Bit_Wait    NOP
            DECFSZ  Delay_Count , f
            GOTO    Bit_Wait
            RETURN

;End of serial routines

		end

