;**********************************************************************
;
; Tutorial 12.2RL - Manchester receive routine, four byte packets.
; display via LCD
;
;**********************************************************************

	list      p=16f628

	__CONFIG 0x3D18

	#include <p16f628.inc>


;***** CONSTANT VALUES
EXPIRE_TIMER	EQU 		0x12
LATCH_MASK	EQU 		0xff
VALID_BIT	EQU 		4
hdrcntmin	EQU		0x0c ; minimum number of header bits to receive
hdrcntmax	EQU		0x10 ; maximum number of header bits to receive

; decoder time tolerances are set at : 0.5T, 1.5T, 2.5T
; measured in 9xinstr.time (9usec) counts

T		EQU		.39 ; half frame 350 usec (= T * 9 usec)

min_t		EQU		T/2  ; half frame (T) minimum time
min_2t		EQU		3*T/2 ; half frame (T) maximum time and full frame (2T) minimum time
max_2t		EQU		5*T/2 ; full frame (2T) maximum time

packet_len	EQU		5 ; packet length, check var. alloc!
mrx_packet_len	EQU		5 ; packet length, check var. alloc!


; input port bit
#define 	RXBIT 		PORTA, 7

; normal decoder logic input
#define 	SKL 		btfsc
#define 	SKH 		btfss

; inverse decoder logic input
;#define 	SKL 		btfss
;#define 	SKH 		btfsc

;***** FLAGS
#define 	IF_SHORT 	flags, 0
#define 	FIRST_HALF 	flags, 1
#define 	HEADER 		flags, 2
#define 	VALID 		flags, 7

if_short_val	EQU		1 ; bit value of IF_SHORT flag
first_half_val	EQU		2 ; bit value of FIRST_HALF flag


;LCD definitions
LCD_PORT	Equ	PORTB
LCD_TRIS	Equ	TRISB
LCD_RS		Equ	0x04			;LCD handshake lines
LCD_RW		Equ	0x06
LCD_E		Equ	0x07

;***** VARIABLE DEFINITIONS

	cblock 	0x20 			;start of general purpose registers

		bitcnt		
		tmrval		 	; timer value
		bt		 	; receive byte buffer
		flags		 	; decode logic status
		btcnt		 	; byte counter

		mrx_buffer 		; receive address buffer
		mrx_buffer1		; receive packet buffer
		mtx_buffer2
		mtx_buffer3
		mtx_buffer4
		mrx_bsum		; receive buffer, checksum

		count			;used in looping routines
		count1			;used in delay routine
		counta			;used in delay routine
		countb			;used in delay routine
		tmp1			;temporary storage
		tmp2
		templcd			;temp store for 4 bit mode
		templcd2

       		NumL			;Binary inputs for decimal convert routine
        	NumH	

       		TenK			;Decimal outputs from convert routine
        	Thou	
       		Hund	
        	Tens	
       		Ones

	endc

		org	0x0000		; org sets the origin, 0x0000 for the 16F628,
  		goto    main		; go to beginning of program
		nop
		nop
		nop
					; interrupt vector location
		retfie

HEX_Table  	ADDWF   PCL       , f
            	RETLW   0x30
            	RETLW   0x31
            	RETLW   0x32
            	RETLW   0x33
            	RETLW   0x34
            	RETLW   0x35
            	RETLW   0x36
            	RETLW   0x37
            	RETLW   0x38
            	RETLW   0x39
            	RETLW   0x41
            	RETLW   0x42
            	RETLW   0x43
            	RETLW   0x44
            	RETLW   0x45
            	RETLW   0x46

main
		;HARDWARE INIT
		movlw 	7
		movwf 	CMCON

SetPorts	bsf 	STATUS,		RP0	; select bank 1
		movlw	0x00			; make all pins outputs
		movwf	LCD_TRIS
		movlw 	0x80 			; A7=input
		movwf 	TRISA
		bcf 	STATUS,		RP0	; select bank 0

		call	LCD_Init		; setup LCD

loop		

		call 	mrx_receive		; get packet
		andlw 	0xff
		bnz 	loop			; reject if error code
		call 	mrx_chk_buf		; test the checksum value
		andlw 	0xff
		bnz 	loop			; reject packet if checksum fails

		movlw	0xAA			; check that packet address is correct (0xAA)
		subwf 	(mrx_buffer), W
		bnz 	loop			; else reject packet

		movf	mrx_buffer+1, W		; get data byte byte
		call	LCD_Char		; and display it
		movf	mrx_buffer+2, W		; get data byte byte
		call	LCD_Char		; and display it
		movf	mrx_buffer+3, W		; get data byte byte
		call	LCD_Char		; and display it
		movf	mrx_buffer+4, W		; get data byte byte
		call	LCD_Char		; and display it

		goto 	loop			; just display the bytes

;*************** receive routines *************************

;    Author:        el@jap.hu
;                   http://jap.hu/electronic/


mrx_receive	; receive a full manchester-encoded packet
s3		; set flags: first_half=1, if_short=0
		bsf 	FIRST_HALF
s4		bcf 	IF_SHORT

s5		; init before the received packet

		; set FSR to buffer start
		movlw 	mrx_buffer
		movwf 	FSR
		; set byte counter
		movlw 	(packet_len+1) ; bytes / packet
		movwf 	btcnt
		; set header receive mode
		bsf 	HEADER
		clrf 	bitcnt ; counting bit1-s in this mode

s2		; wait for a pulse
		SKH 	RXBIT
		goto 	s2

s6		; wait for end of (short) pulse up to min_2t
		clrf 	tmrval
s6_w		SKH 	RXBIT
		goto 	s7 		; goto s7 at end of pulse

		incf 	tmrval, F
		nop
		movlw 	min_2t
		subwf 	tmrval, W
		btfss 	STATUS, C
		goto 	s6_w

		; timeout, exit
		retlw 	1 		; illegal startbit

s7		; start timer
		clrf 	tmrval

s8		; if (if_short & rxbit) goto s9
		; if (!if_short & !rxbit) goto s9
		; goto s10

		btfsc 	IF_SHORT
		; if_short = 1
		goto 	s8_ss1

s8_ss0		; if_short = 0
		SKL 	RXBIT
		goto 	s10 		; rxbit = 1, goto s10

s9_ss0		; if (timer > max_2t) exit - else goto s8
		movlw 	max_2t
		subwf 	tmrval, W
		btfsc 	STATUS, C
		retlw 	2 		; signal too long

		incf 	tmrval, F
		goto 	s8_ss0

s8_ss1		; if_short = 1
		SKH 	RXBIT
		goto 	s10 ; rxbit = 0, goto s10

s9_ss1		; if (timer > max_2t) exit - else goto s8
		movlw 	max_2t
		subwf 	tmrval, W
		btfsc 	STATUS, C
		retlw 	2 		; signal too long

		incf 	tmrval, F
		goto 	s8_ss1

s10		; invert if_short
		movlw 	if_short_val
		xorwf 	flags, F

s11		; if (timer < min_t) exit
		movlw 	min_t
		subwf 	tmrval, W
		btfss 	STATUS, C
		retlw 3 		; signal too short

s12		; if (timer < min_2t) goto s14
		movlw 	min_2t
		subwf 	tmrval, W
		btfss 	STATUS, C
		goto 	s14

s13		; if (first_half = 0) goto s16 - else exit
		btfss 	FIRST_HALF
		goto 	s16
		retlw 	4 		; no mid-frame transition/out of sync

s14		; invert first_half
		movlw 	first_half_val
		xorwf 	flags, F

s15		; if (first_half = 1) goto 7
		btfsc 	FIRST_HALF
		goto 	s7

s16		; if_short is a decoded bit. Handle here
		btfss 	HEADER
		goto 	s16_not_header

		; header receiving mode
		btfss 	IF_SHORT
		goto 	s16_header_end

		; header bit is 1
		btfss 	bitcnt, 4 ; inc up to 16
		incf 	bitcnt, F  ; 16 is enough...

#ifdef NOMAXHDR
		; test for max header length
		movlw 	hdrcntmax
		subwf 	bitcnt, W
		btfss 	STATUS, C
#endif
		goto 	s7 ; loop back
		retlw 	9 		; header too long

s16_header_end	; header ends indicated by a 0
		bcf 	HEADER

		; test for min header length
		movlw 	hdrcntmin
		subwf 	bitcnt, W
		btfss 	STATUS, C
		retlw 	0x0a 		; header too short

next_byte	movlw 	0x0a
		movwf 	bitcnt
		goto 	s7 ; loop back

s16_not_header	; receiving bytes
		decf 	bitcnt, F
		bz 	s16_s4 ; if (bitcnt = 0) check for a byte-sep 1

		; if (bitcnt = 1) check for a byte-separator 0
		movlw 	1
		xorwf 	bitcnt, W
		bnz 	s16_s2

		; test for a byte separator 1
		btfsc 	IF_SHORT
		goto 	s7
		retlw 	7 		; byte-ending 1 not present

s16_s2		; bit is data

		rrf 	flags, W
		rlf 	bt, F
		goto 	s7

s16_s4		; check for a byte-separator 0
		btfsc 	IF_SHORT
		retlw 	8 		; byte-ending 0 not present

		; OK, received byte is sane, store in buffer
		movf 	bt, W
		movwf 	INDF
		incf 	FSR, F

		decfsz 	btcnt, F
		goto 	next_byte

		retlw 	0 		; OK, buffer received

		; buffer checking is not done automatically!
		; if returned value is 0, call mrx_chk_buf to check
		;

mrx_chk_buf	; check buffer sanity by chksum
		movlw 	mrx_buffer
		movwf 	FSR
		movlw 	(packet_len+1) ; number of bytes with the chksum byte
		movwf 	btcnt
		movlw 	0xff
		movwf 	bt ; used as sum register

chk0		movf 	INDF, W
		; fast CRC-8 algorithm with poly x^8+x^5+x^4+1
		; executes in 23 cycles per update
		xorwf	bt,f
		clrw
		btfsc	bt,7
		xorlw	0x7a
		btfsc	bt,6
		xorlw	0x3d
		btfsc	bt,5
		xorlw	0x86
		btfsc	bt,4
		xorlw	0x43
		btfsc	bt,3
		xorlw	0xb9
		btfsc	bt,2
		xorlw	0xc4
		btfsc	bt,1
		xorlw	0x62
		btfsc	bt,0
		xorlw	0x31
		movwf	bt

		incf 	FSR, F
		decfsz 	btcnt, F
		goto 	chk0
		; correct checksum must be zero
		movf 	bt, W
		bnz 	chk_err
		retlw 	0 		; result is in Z
chk_err		retlw 	0x0c 		; checksum error

mrx_init	return

;LCD routines

;Initialise LCD
LCD_Init	call 	LCD_Busy		;wait for LCD to settle

		movlw	0x20			;Set 4 bit mode
		call	LCD_Cmd

		movlw	0x28			;Set display shift
		call	LCD_Cmd

		movlw	0x06			;Set display character mode
		call	LCD_Cmd

		movlw	0x0c			;Set display on/off and cursor command
		call	LCD_Cmd			;Set cursor off

		call	LCD_Clr			;clear display

		retlw	0x00

; command set routine
LCD_Cmd		movwf	templcd
		swapf	templcd,	w	;send upper nibble
		andlw	0x0f			;clear upper 4 bits of W
		movwf	LCD_PORT
		bcf	LCD_PORT, LCD_RS	;RS line to 1
		call	Pulse_e			;Pulse the E line high

		movf	templcd,	w	;send lower nibble
		andlw	0x0f			;clear upper 4 bits of W
		movwf	LCD_PORT
		bcf	LCD_PORT, LCD_RS	;RS line to 1
		call	Pulse_e			;Pulse the E line high
		call 	LCD_Busy
		retlw	0x00

LCD_CharD	addlw	0x30			;add 0x30 to convert to ASCII
LCD_Char	movwf	templcd
		swapf	templcd,	w	;send upper nibble
		andlw	0x0f			;clear upper 4 bits of W
		movwf	LCD_PORT
		bsf	LCD_PORT, LCD_RS	;RS line to 1
		call	Pulse_e			;Pulse the E line high

		movf	templcd,	w	;send lower nibble
		andlw	0x0f			;clear upper 4 bits of W
		movwf	LCD_PORT
		bsf	LCD_PORT, LCD_RS	;RS line to 1
		call	Pulse_e			;Pulse the E line high
		call 	LCD_Busy
		retlw	0x00

LCD_Line1	movlw	0x80			;move to 1st row, first column
		call	LCD_Cmd
		retlw	0x00

LCD_Line2	movlw	0xc0			;move to 2nd row, first column
		call	LCD_Cmd
		retlw	0x00

LCD_Line1W	addlw	0x80			;move to 1st row, column W
		call	LCD_Cmd
		retlw	0x00

LCD_Line2W	addlw	0xc0			;move to 2nd row, column W
		call	LCD_Cmd
		retlw	0x00

LCD_CurOn	movlw	0x0d			;Set display on/off and cursor command
		call	LCD_Cmd
		retlw	0x00

LCD_CurOff	movlw	0x0c			;Set display on/off and cursor command
		call	LCD_Cmd
		retlw	0x00

LCD_Clr		movlw	0x01			;Clear display
		call	LCD_Cmd
		retlw	0x00

LCD_HEX		movwf	tmp1
		swapf	tmp1,	w
		andlw	0x0f
		call	HEX_Table
		call	LCD_Char
		movf	tmp1, w
		andlw	0x0f
		call	HEX_Table
		call	LCD_Char
		retlw	0x00

Delay255	movlw	0xff			;delay 255 mS
		goto	d0
Delay100	movlw	d'100'			;delay 100mS
		goto	d0
Delay50		movlw	d'50'			;delay 50mS
		goto	d0
Delay20		movlw	d'20'			;delay 20mS
		goto	d0
Delay5		movlw	0x05			;delay 5.000 ms (4 MHz clock)
d0		movwf	count1
d1		movlw	0xC7			;delay 1mS
		movwf	counta
		movlw	0x01
		movwf	countb
Delay_0
		decfsz	counta, f
		goto	$+2
		decfsz	countb, f
		goto	Delay_0

		decfsz	count1	,f
		goto	d1
		retlw	0x00

Pulse_e		bsf	LCD_PORT, LCD_E
		nop
		bcf	LCD_PORT, LCD_E
		retlw	0x00

LCD_Busy
		bsf	STATUS,	RP0		;set bank 1
		movlw	0x0f			;set Port for input
		movwf	LCD_TRIS
		bcf	STATUS,	RP0		;set bank 0
		bcf	LCD_PORT, LCD_RS	;set LCD for command mode
		bsf	LCD_PORT, LCD_RW	;setup to read busy flag
		bsf	LCD_PORT, LCD_E
		swapf	LCD_PORT, w		;read upper nibble (busy flag)
		bcf	LCD_PORT, LCD_E		
		movwf	templcd2 
		bsf	LCD_PORT, LCD_E		;dummy read of lower nibble
		bcf	LCD_PORT, LCD_E
		btfsc	templcd2, 7		;check busy flag, high = busy
		goto	LCD_Busy		;if busy check again
		bcf	LCD_PORT, LCD_RW
		bsf	STATUS,	RP0		;set bank 1
		movlw	0x00			;set Port for output
		movwf	LCD_TRIS
		bcf	STATUS,	RP0		;set bank 0
		return

;end of LCD routines



;This routine downloaded from http://www.piclist.com
Convert:                        ; Takes number in NumH:NumL
                                ; Returns decimal in
                                ; TenK:Thou:Hund:Tens:Ones
        swapf   NumH, w
	iorlw	B'11110000'
        movwf   Thou
        addwf   Thou,f
        addlw   0XE2
        movwf   Hund
        addlw   0X32
        movwf   Ones

        movf    NumH,w
        andlw   0X0F
        addwf   Hund,f
        addwf   Hund,f
        addwf   Ones,f
        addlw   0XE9
        movwf   Tens
        addwf   Tens,f
        addwf   Tens,f

        swapf   NumL,w
        andlw   0X0F
        addwf   Tens,f
        addwf   Ones,f

        rlf     Tens,f
        rlf     Ones,f
        comf    Ones,f
        rlf     Ones,f

        movf    NumL,w
        andlw   0X0F
        addwf   Ones,f
        rlf     Thou,f

        movlw   0X07
        movwf   TenK

                    ; At this point, the original number is
                    ; equal to
                    ; TenK*10000+Thou*1000+Hund*100+Tens*10+Ones
                    ; if those entities are regarded as two's
                    ; complement binary.  To be precise, all of
                    ; them are negative except TenK.  Now the number
                    ; needs to be normalized, but this can all be
                    ; done with simple byte arithmetic.

        movlw   0X0A                             ; Ten
Lb1:
        addwf   Ones,f
        decf    Tens,f
        btfss   3,0
        goto   Lb1
Lb2:
        addwf   Tens,f
        decf    Hund,f
        btfss   3,0
        goto   Lb2
Lb3:
        addwf   Hund,f
        decf    Thou,f
        btfss   3,0
        goto   Lb3
Lb4:
        addwf   Thou,f
        decf    TenK,f
        btfss   3,0
        goto   Lb4

        retlw	0x00


		end

