;**********************************************************************
;
; Tutorial 12.5Ta - Manchester transmit routine, four byte packets.
; A2D twin channel transmitter 16F876 at 20MHz.
;
;**********************************************************************
	list		p=16f876
	__CONFIG    	0x393A


#include <p16F876.inc>

;#define 		TXINV
;invert TX output so inactive is HIGH?




packet_len	EQU 	5		;packet length 1 + 1 address byte

;TX pin is portb, 6

#ifndef 	TXINV
#define 	TXLOW 	BCF	PORTB, 6
#define 	TXHIGH 	BSF	PORTB, 6
#else
#define 	TXLOW 	BSF	PORTB, 6
#define 	TXHIGH	BCF	PORTB, 6
#endif

SWPORT		Equ	PORTC			;set constant SWPORT = 'PORTC
SW1		Equ	7			;set constants for the switches
SW2		Equ	6
SW3		Equ	5
SW4		Equ	4


;***** VARIABLE DEFINITIONS

	cblock 	0x20 			;start of general purpose registers

		ncnt
		ncnt1
		bt
		sum
		mtx_buffer
		mtx_buffer1
		mtx_buffer2
		mtx_buffer3
		mtx_buffer4
		mtx_delay		; half_frame delay

		count1			;used in delay routine
		count2
		counta			;used in delay routine
		countb			;used in delay routine

		NumL			;used in A2D routines
		NumH	
	endc



		org	0x0000
  		goto 	main
		nop
		nop
		nop
		retfie


main		; program starts here

		BANKSEL ADCON1			;disable A2D
    		movlw   0x06
    		movwf   ADCON1
    		BANKSEL PORTA

		bsf 	STATUS,		RP0	;select bank 1
		movlw 	0x00
		movwf 	TRISB
		movlw	0xFF			;set PortC to all inputs
		movwf 	TRISC



		bcf 	STATUS,		RP0	;select bank 0
		clrf 	PORTC
		clrf 	PORTB
		call 	mtx_init
		movlw	0xAA			;set address byte to 0xAA
		movwf	mtx_buffer
		

Loop		call	Init_ADC0		;set for first channel
		call	Delay50			;delay to give 10 readings per second
		;read and store first input
		call	Read_ADC		;read analogue input
		movf	NumH, W
		movwf	mtx_buffer+1
		movf	NumL, W
		movwf	mtx_buffer+2

		call	Init_ADC1		;set for second channel
		call	Delay50			;delay to give 10 readings per second

		;read and store second input
		call	Read_ADC		;read analogue input
		movf	NumH, W
		movwf	mtx_buffer+3
		movf	NumL, W
		movwf	mtx_buffer+4

		call	mtx_send		;send via wireless module
		
		goto	Loop

Init_ADC0					;set for AN0
; Set ADCON0
    		movlw   b'10000001'
    		movwf   ADCON0
; Set ADCON1
    		BANKSEL ADCON1
    		movlw   b'10000101'
    		movwf   ADCON1
    		BANKSEL ADCON0
		return

Init_ADC1					;set for AD1
; Set ADCON0
    		movlw   b'10001001'
    		movwf   ADCON0
; Set ADCON1
    		BANKSEL ADCON1
    		movlw   b'10000101'
    		movwf   ADCON1
    		BANKSEL ADCON0
		return

Read_ADC
    		bsf	ADCON0, GO_DONE		;initiate conversion
    		btfsc   ADCON0, GO_DONE
    		goto    $-1			;wait for ADC to finish

    		movf    ADRESH,W
    		andlw   0x03
    		movwf   NumH
    		BANKSEL ADRESL
    		movf    ADRESL,W
    		BANKSEL	ADRESH
		movwf	NumL			;return result in NumL and NumH
		return


;Manchester subroutines

;    Author:        el@jap.hu
;                   http://jap.hu/electronic/

mtx_init	movlw 	.115 		; 350 usec
		movwf 	mtx_delay
		return
		;

mtx_send
					; send out buffer
outbuf		movlw 	0x14 		; 20xbit1, 1xbit0

header		movwf 	count2
head0		call 	bit1
		decfsz 	count2,F
		goto 	head0
		call 	bit0

		movlw 	mtx_buffer
		movwf 	FSR
		movlw 	packet_len
		movwf 	count1
		movlw 	0xff
		movwf 	sum
		;
outbu0		movf 	INDF,W
		call 	update_sum
		movf 	INDF,W
		call 	outbyte
		incf 	FSR,F
		decfsz 	count1,F
		goto 	outbu0
		movf 	sum,W
		call 	outbyte
					; buffer is sent
		return

update_sum	; fast CRC-8 algorithm with poly x^8+x^5+x^4+1
		; executes in 23 cycles per update

		xorwf	sum,f
		clrw
		btfsc	sum,7
		xorlw	0x7a
		btfsc	sum,6
		xorlw	0x3d
		btfsc	sum,5
		xorlw	0x86
		btfsc	sum,4
		xorlw	0x43
		btfsc	sum,3
		xorlw	0xb9
		btfsc	sum,2
		xorlw	0xc4
		btfsc	sum,1
		xorlw	0x62
		btfsc	sum,0
		xorlw	0x31
		movwf	sum
		return

outbyte		movwf 	bt
		movlw 	8
		movwf 	count2
outby0		rlf 	bt,F
		btfsc 	STATUS,C
		goto 	outby1
		call 	bit0
		goto 	outby2
outby1		call 	bit1
outby2		decfsz 	count2,F
		goto 	outby0
		;
		call 	bit1
		; and 	bit0 - falls through to bit0 subroutine

;send a bit0

bit0		TXHIGH 			; HIGH


		call	mtx_bitdel	; bit time delay

		TXLOW 			; to LOW transition

		call	mtx_bitdel	; bit time delay
		return

;send a bit1

bit1		TXLOW 			; LOW

		call	mtx_bitdel	; bit time delay

		TXHIGH 			; to HIGH transition

		call	mtx_bitdel	; bit time delay
		return

; bit delay modified by NG

mtx_bitdel	movlw	D'3'
		movwf	ncnt
		movlw	D'68'
		movwf	ncnt1
loopD		decfsz	ncnt1,f
		goto	loopD
		decfsz	ncnt,f
		goto	loopD
		return

; end of Manchester routines

;Delay routines

Delay255	movlw	0xff		;delay 255 mS
		goto	d0
Delay100	movlw	d'100'		;delay 100mS
		goto	d0
Delay50		movlw	d'50'		;delay 50mS
		goto	d0
Delay20		movlw	d'20'		;delay 20mS
		goto	d0
Delay5		movlw	0x05		;delay 5.000 ms (20 MHz clock)
d0		movwf	count1
d1		movlw	0xE7
		movwf	counta
		movlw	0x04
		movwf	countb
Delay_0		decfsz	counta, f
		goto	$+2
		decfsz	countb, f
		goto	Delay_0

		decfsz	count1	,f
		goto	d1
		retlw	0x00

;end of Delay routines


		end
